

window.CodeMirror = require('codemirror');
require('codemirror/mode/javascript/javascript');
require('codemirror/addon/hint/show-hint');
require('codemirror/addon/hint/javascript-hint');
require('codemirror/mode/htmlmixed/htmlmixed');
require('codemirror/addon/hint/html-hint');
require('codemirror/mode/css/css');
require('codemirror/addon/hint/css-hint');
require('codemirror/mode/jsx/jsx');

require('codemirror/addon/dialog/dialog');
require('codemirror/addon/search/search');
require('codemirror/addon/search/searchcursor');
require('codemirror/addon/search/jump-to-line');
require('codemirror/addon/selection/mark-selection');



CodeMirror.defineOption('showInvisibles', false, (cm, val, prev) => {
  let Count = 0;
  const Maximum = cm.getOption('maxInvisibles') || 16;
  
  if(prev === CodeMirror.Init) {
    prev = false;
  }
  
  if(prev && !val) {
    cm.removeOverlay('invisibles');
    return rm();
  }
  
  if(!prev && val) {
    add(Maximum);
    
    cm.addOverlay({
      name: 'invisibles',
      token: function nextToken(stream) {
        let spaces = 0;
        let peek = stream.peek() === ' ';
        
        if(peek) {
          while(peek && spaces < Maximum) {
            ++spaces;
            stream.next();
            peek = stream.peek() === ' ';
          }
          
          let ret = 'whitespace whitespace-' + spaces;
          if(spaces === Maximum) {
            ret += ' whitespace-rand-' + Count++;
          }
          return ret;
        }
        
        while(!stream.eol() && !peek) {
          stream.next();
          peek = stream.peek() === ' ';
        }
        return 'cm-eol';
      }
    });
  }
});
    
function add(max) {
  const classBase = '.CodeMirror .cm-whitespace-';
  const spaceChar = '\u00B7';
  const tabChar = '\u2192';
  const style = document.createElement('style');
  
  style.setAttribute('data-name', 'js-show-invisibles');
  
  let rules = '';
  let spaceChars = '';
  
  for(let i = 1; i <= max; ++i) {
    spaceChars += spaceChar;
    rules += classBase + i + `:not([class*="cm-trailing-space-"])::before { content: "${spaceChars}";}\n\n`;
  }
  
  const gfmRules = `[class*=cm-trailing-space]::before{content: "${spaceChar}";}`;
  
  const tabRules = `span.cm-tab::before{content: "${tabChar}";position:absolute;}`;
  
  style.textContent = `${getStyle()}\n${getEOL()}\n${rules}\n${gfmRules}\n${tabRules}`;
  document.head.appendChild(style);
}

function rm() {
  const style = document.querySelector('[data-name="js-show-invisibles"]');
  document.head.removeChild(style);
}

function getStyle() {
  return `
.cm-whitespace::before {
  position: absolute;
  pointer-events: none;
  color: #404F7D;
}`;
}
    
function getEOL() {
  return `
.CodeMirror-code > div > pre > span::after, .CodeMirror-line > span::after {
   pointer-events: none;
   color: #404F7D;
   content: "¬"
}

.CodeMirror-code > div:last-child > pre > span::after {',
   display: none;
}
`;
}
